/*
 * 	g2cap.c
 *
 *	G2 Capstone
 *	Created by: Nigel Himmelreich
 *	ECE 492 - Winter 2015 - University of Alberta
 *
 * DE0 Nano Board Code
 * Provides NFC and Bluetooth connectivity on the DEO Nano board
 *
 */

#include <stdio.h>
#include <stdint.h>
#include "sys\alt_irq.h"
#include "includes.h"
#include "system.h"
#include "altera_up_avalon_rs232.h"
#include "altera_up_avalon_rs232_regs.h"
#include "btSerial.h"
#include "I2C.h"
#include "pn532.h"

/* Definition of Task Stacks */
#define   TASK_STACKSIZE       2048
OS_STK    task_nfc_stk[TASK_STACKSIZE];
OS_STK    task_bt_stk[TASK_STACKSIZE];

/* Definition of Task Priorities */
#define TASK_NFC_PRIORITY      	1
#define TASK_BT_PRIORITY      	2

/* Definition of Queue */
#define IN_QUEUE_SIZE			10
void * in_data_storage[IN_QUEUE_SIZE];
OS_EVENT * in_data_queue;

/* NFC IRQ Semaphore*/
OS_EVENT * NFC_IRQ_Sem;

/* NFC IRQ */
static void isr_nfc_irq(void* context, alt_u32 id) {
	//Release semaphore and clear IRQ
	IOWR_ALTERA_AVALON_PIO_EDGE_CAP(NFC_IRQ_BASE, NFC_IRQ_BIT_CLEARING_EDGE_REGISTER);
	OSSemPost(NFC_IRQ_Sem);
}

/*
 * Task for the NFC functionality
 *
 * Inits the NFC module and reads data as needed.
 */
void task_nfc(void* pdata) {
	//Init the connection to the NFC unit (Adafuit PN532) over I2C
	INT8U err = OS_NO_ERR;
	alt_u8 targets = 0;
	TargetData listTargets[PN532_MAX_TARGETS];
	int i = 0;

	//Start
	printf("Start NFC task\n");
	IOWR_ALTERA_AVALON_PIO_DATA(GREEN_LEDS_L_BASE, 0x0F);

	//NFC Semaphore for the IRQ
	NFC_IRQ_Sem = OSSemCreate(0);
	if (NFC_IRQ_Sem == NULL) {
		printf("Error: NFC IRQ Semaphore failed\n");
	} else {
		printf("Success: NFC IRQ Semaphore completed\n");
	}

	while(1) {
		//Reset the NFC board
		nfcBegin();
		OSTimeDlyHMSM(0, 0, PN532_RESET_DELAY, 0);

		//Configure the NFC reader to accept type B cards, with timeout, and IRQ on
		samConfiguration(PN532_NORMAL_MODE, PN532_VCARD_TIMEOUT, PN532_VCARD_IRQ_ON);
		OSSemPend(NFC_IRQ_Sem, 0, &err);
		if (!checkAck() || err != OS_NO_ERR) {
			printf("Error: samConfiguration during init ack failed\n");
		} else {
			printf("Success: samConfiguration of init ack completed\n");
		}
		OSSemPend(NFC_IRQ_Sem, 0, &err);
		if(!checkGeneralSuccess() || err != OS_NO_ERR) {
			printf("Error: samConfiguration during init response failed\n");
		} else {
			printf("Success: samConfiguration during init response completed\n");
		}

		//Set RF field to full power for use with the induction constantly
		commandRFFieldConfig();
		OSSemPend(NFC_IRQ_Sem, 0, &err);
		if (!checkAck() || err != OS_NO_ERR) {
			printf("Error: RF Config failed\n");
		} else {
			printf("Success: RF Config completed\n");
		}
		OSSemPend(NFC_IRQ_Sem, 0, &err);
		if (!checkGeneralSuccess() || err != OS_NO_ERR) {
			printf("Error: RF Config response failed\n");
		} else {
			printf("Success: RF Config response completed\n");
		}
		printf("Looking for targets\n");

		//Setup to look for a target
		inListPassiveTarget(PN532_MAX_TARGETS, PN532_ISO_B);
		OSSemPend(NFC_IRQ_Sem, 0, &err);
		if (!checkAck() || err != OS_NO_ERR) {
			continue;
		}
		printf("Wait for NFC Card\n");
		IOWR_ALTERA_AVALON_PIO_DATA(GREEN_LEDS_L_BASE, 0x03);

		//Wait for NFC card for the interrupt to be triggered
		OSSemPend(NFC_IRQ_Sem, 0, &err);
		printf("Cleared semaphore\n");
		if (!parsePassiveTargetData(&targets, listTargets) || err != OS_NO_ERR) {
			continue;
		}
		printf("Found target(s), check each one.\n");
		IOWR_ALTERA_AVALON_PIO_DATA(GREEN_LEDS_L_BASE, 0x05);

		//For each target, loop through checking each info
		while (targets > 0) {
			printf("numTargets > 0.\n");
			targets--;

			printf("Target number: %d\n", listTargets[targets].targetNumber);
			printf("Target idLen: %d\n", listTargets[targets].idLen);

			for (i=0; (i < listTargets[targets].idLen); i++) {
				printf("0x%X ", listTargets[targets].id[i]);
			}
			printf("\n");

			char c[10];
			sprintf(c, "%d %d\n", listTargets[targets].id[0], listTargets[targets].id[1]);

			//Post data (ID # 2 values) to the queue
			OSQPost(in_data_queue, (void *)c);
		}//End-while

		printf("Through all targets.\n");
		OSTimeDlyHMSM(0, 0, 0, 1000);
	}//Loop forever
}

/*
 * Task for the Bluetooth functionality
 *
 * Inits the Bluetooth module and sends commands as needed.
 */
void task_bt(void* pdata) {
	printf("Start bt task\n");
	//Init the connection to the Bluetooth unit (MOD-BT) over serial
	alt_up_rs232_dev * serial_bt = initDevice(UART_0_NAME);

	setupModule(serial_bt, "AT+BTLNM=\"George\"\r", "AT+BTPIN=\"2001\"\r");

	//Delay start of data from connection (adjust or remove with testing)
	IOWR_ALTERA_AVALON_PIO_DATA(GREEN_LEDS_R_BASE, 0x0F);
	printf("Wait for delay\n");
	OSTimeDlyHMSM(0, 0, 5, 0);
	printf("End of delay\n");
	IOWR_ALTERA_AVALON_PIO_DATA(GREEN_LEDS_R_BASE, 0x03);

	char * current;
	char * saved;
	INT8U err = OS_NO_ERR;
	bool firstData = true;
	void * result;

	//Loop forever through receiving values and sending data
	while (1) {
		//If first time getting data is true, pend until first data is received from NFC
		if (firstData == true) {
			result = OSQPend(in_data_queue, 0, &err);
			printf("First data\n");
			firstData = false;
			current = (char*) result;
			saved = current;
		} else {
			//Else, wait to see if new data is there
			result = OSQPend(in_data_queue, 1000, &err);
			current = (char*) result;
		}

		if (err == OS_ERR_TIMEOUT || result == NULL) {
			//If timed out with no new result
			printf("Timed out\n");
			IOWR_ALTERA_AVALON_PIO_DATA(GREEN_LEDS_R_BASE, 0x09);
			multiWrite(saved, serial_bt);
		} else if (err != OS_NO_ERR) {
			printf("Error: Queue pend failed\n");
		} else {
			//New data, write and save it for next times until new data arrives
			IOWR_ALTERA_AVALON_PIO_DATA(GREEN_LEDS_R_BASE, 0x05);
			multiWrite(current, serial_bt);
			saved = current;
		}
		OSTimeDlyHMSM(0, 0, 1, 0);
	}//End while
}

/*
 * Main function to start applicable tasks
 */
int main(void) {
	printf("Hello\n");

	//Set LEDS
	IOWR_ALTERA_AVALON_PIO_DATA(GREEN_LEDS_R_BASE, 0x00);
	IOWR_ALTERA_AVALON_PIO_DATA(GREEN_LEDS_L_BASE, 0x00);

	//Set the NFC IRQ
	int isr;
	isr = alt_irq_register(NFC_IRQ_IRQ, NULL, (void*)isr_nfc_irq);
	IOWR_ALTERA_AVALON_PIO_IRQ_MASK(NFC_IRQ_BASE, NFC_IRQ_CAPTURE);
	printf("ISR Register init: %d\n", isr);

	//Generate a queue for passing data from NFC to Bluetooth
	in_data_queue = OSQCreate(in_data_storage, IN_QUEUE_SIZE);

	//Create a task for the Bluetooth functionality
	OSTaskCreateExt(task_bt, NULL, (void *)&task_bt_stk[TASK_STACKSIZE-1], TASK_BT_PRIORITY, TASK_BT_PRIORITY, task_bt_stk, TASK_STACKSIZE, NULL, 0);

	//Create a task for the NFC functionality
	OSTaskCreateExt(task_nfc, NULL, (void *)&task_nfc_stk[TASK_STACKSIZE-1], TASK_NFC_PRIORITY, TASK_NFC_PRIORITY, task_nfc_stk, TASK_STACKSIZE, NULL, 0);

	OSStart();
	return 0;
}
