/*
 * pn532.c
 *
 *  Created on: 2014-02-23
 *      Author: Daniel
 */

/*
 * Modified by: Nigel Himmelreich for ECE 492 - Winter 2015 - Group 2
 *
 * Added additional methods and fixed for use with DE0 Nano and Adafruit Shield
 * Add: nfcBegin; commandReadDataBlock; parseReadDataBlock; releaseCard, parseCard,
 * 		setParameters, commandDataSend, selectCard, parseSelectCard, commandRFConfig,
 * 		commandRFRtyConfig, commandInATR, commandInPSL
 */

#include "pn532.h"

//Frame and data buffers for sending/receiving data from PN532
alt_u8 frameBuffer[PN532_MAX_FRAME_SIZE];
alt_u8 dataBuffer[PN532_MAX_DATA_SIZE];
alt_u8 lastCommand;

//ACK and Firmware frame to compare to for accurate values
const alt_u8 ackFrame[] = {0x00, 0x00, 0xFF, 0x00, 0xFF, 0x00};
const alt_u8 pn532Firmware[] = {0x32, 0x01, 0x06, 0x07};

//Internal
bool readStatusReady();
bool readResponse();
void sendCommand(int cmdLen);
void packFrame(alt_u8 cmdLen);
void sendFrame(alt_u8 frmLen);

/*
 * Inits the PN532 by resetting it (HIGH, LOW, wait 400ms, HIGH, wait)
 */
void nfcBegin() {
	IOWR_ALTERA_AVALON_PIO_DATA(PN532_RESET_BASE, 1);
	IOWR_ALTERA_AVALON_PIO_DATA(PN532_RESET_BASE, 0);
	usleep(RESET_DELAY);
	IOWR_ALTERA_AVALON_PIO_DATA(PN532_RESET_BASE, 1);
}

/*
 * Send a command to the PN532 to initiate the input of data
 */
void commandReadDataBlock(alt_u8 targetNumber, alt_u8 sendLen, alt_u8 * data) {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);
	dataBuffer[0] = PN532_COMMAND_INDATAEXCHANGE;
	dataBuffer[1] = 1;

	dataBuffer[2] = 0x00;
	dataBuffer[3] = 0xA4;
	dataBuffer[4] = 0x04;
	dataBuffer[5] = 0x00;
	dataBuffer[6] = 0x07;
	dataBuffer[7] = 0xD2;
	dataBuffer[8] = 0x76;
	dataBuffer[9] = 0x00;
	dataBuffer[10] = 0x00;
	dataBuffer[11] = 0x85;
	dataBuffer[12] = 0x01;
	dataBuffer[13] = 0x01;
	dataBuffer[14] = 0x00;

	lastCommand = PN532_COMMAND_INDATAEXCHANGE;
	sendCommand(15);
}

/*
 * Parses the data block from the read data exchange
 * Gets the status byte from the PN532
 *
 * TODO: Extend to all data being received
 */
alt_u8 parseReadDataBlock(alt_u8 * data) {
	memset(frameBuffer, 0, PN532_MAX_FRAME_SIZE);

	//This one gets the whole length specified (no need for readmore/stop)
	if (!readResponse())
		return false;

	alt_u8 testDataBuffer[PN532_MAX_DATA_SIZE];
	memset(testDataBuffer, 0, PN532_MAX_DATA_SIZE);
	memcpy(testDataBuffer, dataBuffer, PN532_MAX_DATA_SIZE);

	data[0] = dataBuffer[2]; //Currently only sends out status byte

	return true;
}

/*
 * Selects the card as specified by the target number for use with commands
 */
void selectCard(alt_u8 targetNumber) {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);
	dataBuffer[0] = PN532_COMMAND_INSELECT;
	dataBuffer[1] = targetNumber;

	lastCommand = PN532_COMMAND_INSELECT;
	sendCommand(2);
}

/*
 * Releases the card given by the target number
 */
void releaseCard(alt_u8 targetNumber) {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);
	dataBuffer[0] = PN532_COMMAND_INRELEASE;
	dataBuffer[1] = targetNumber;

	lastCommand = PN532_COMMAND_INRELEASE;
	sendCommand(2);
}

/*
 * Parses the response from a card on the select/release commands
 */
bool parseCard() {
	if (!readResponse())
		return false;

	if (dataBuffer[2] != PN532_STATUS_OK) {
		printf("Error: Status not okay, card not selected/released\n");
		return false;
	}
	return true;
}

/*
 * Sets the parameters in the PN532 as defined in user manual
 *
 * Sets the DID/NAD bytes
 * 		(NAD is after DID(aka CID) and part of PCB in the ISO 14443-4 standard)
 * 		When DID (aka CID) is used, the values is fixed at 0x01
 * 		The NAD is set by host controller directly for InDataExchange
 */
void setParameters()
{
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);
	dataBuffer[0] = PN532_COMMAND_SETPARAMETERS;
							//bit 0 -> it 7
//	dataBuffer[1] = 0xEC; //0b11101100 //NAD bit used, DID bit used
//	dataBuffer[1] = 0x6C; //NAD bit not used, DID bit used
//	dataBuffer[1] = 0xAC; //NAD bit used, DID bit not used
	dataBuffer[1] = 0x2C; //NAD bit not used, DID bit not used

	lastCommand = PN532_COMMAND_SETPARAMETERS;
	sendCommand(2);
}

/*
 * Sends the given command of the given length to the PN532s
 */
void commandDataSend(alt_u8 command[], alt_u8 len) {
	alt_u8 i;
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);

//	dataBuffer[0] = PN532_COMMAND_INDATAEXCHANGE;
	dataBuffer[0] = PN532_COMMAND_INCOMMUNICATETHRU;
//	dataBuffer[1] = 1;

	for (i=0; i < len; i++) {
//		dataBuffer[i+2] = command[i];
		dataBuffer[i+1] = command[i];
	}

//	lastCommand = PN532_COMMAND_INDATAEXCHANGE;
	lastCommand = PN532_COMMAND_INCOMMUNICATETHRU;
//	sendCommand(len + 2);
	sendCommand(len + 1);
}



/*
 * Configures the RF timing command as defined in the user manual
 * Sets the ATR_RES timeout and non-DEP timeout to be longer
 */
void commandRFConfig() {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);

	dataBuffer[0] = PN532_COMMAND_RFCONFIGURATION;
	dataBuffer[1] = 0x02; //Timings

	dataBuffer[2] = 0x00; //reserved for future use
	dataBuffer[3] = 0x10; //3.28sec
	dataBuffer[4] = 0x10; //3.28sec

	lastCommand = PN532_COMMAND_RFCONFIGURATION;
	sendCommand(5);
}

/*
 * Configures the RF MaxRty Command as defined in the user manual
 * Extends the number of retries in InCommunicateThru in case of error
 */
void commandRFRtyConfig() {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);

	dataBuffer[0] = PN532_COMMAND_RFCONFIGURATION;
	dataBuffer[1] = 0x04; //MaxRtyCom

	dataBuffer[2] = 0x0A;

	lastCommand = PN532_COMMAND_RFCONFIGURATION;
	sendCommand(3);
}

/*
 * Configures the RF Field Command as defined in the user manual
 *
 */
void commandRFFieldConfig() {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);

	dataBuffer[0] = PN532_COMMAND_RFCONFIGURATION;
	dataBuffer[1] = 0x01; //RF Field Config

	dataBuffer[2] = 0x64; //bit 0 -> bit 7; 0b1000 0000

	lastCommand = PN532_COMMAND_RFCONFIGURATION;
	sendCommand(3);
}

/*
 * Configures the RF Field Command as defined in the user manual
 *
 */
void commandRFRetriesConfig() {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);

	dataBuffer[0] = PN532_COMMAND_RFCONFIGURATION;
	dataBuffer[1] = 0x05; //RF Max Retries

	dataBuffer[2] = 0x02; //Passive mode max
	dataBuffer[3] = 0xFF; //Resend max PPS request
	dataBuffer[4] = 0xFF; //Passive activation retry

	lastCommand = PN532_COMMAND_RFCONFIGURATION;
	sendCommand(3);
}

/*
 * Sends the ATR command as defined in the user manual
 */
void commandInATR(alt_u8 targetNumber) {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);

	dataBuffer[0] = PN532_COMMAND_INATR;
	dataBuffer[1] = targetNumber;
	dataBuffer[2] = 0x00; //NFCID3i & Gi not present, sets NFCID3i with random value instead

	lastCommand = PN532_COMMAND_INATR;
	sendCommand(3);
}

/*
 * Sends the PSL command as defined in the user manaul
 */
void commmandInPSL(alt_u8 targetNumber) {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);

	dataBuffer[0] = PN532_COMMAND_INPSL;
	dataBuffer[1] = targetNumber;
	dataBuffer[2] = 0x00;
	dataBuffer[3] = 0x00;

	lastCommand = PN532_COMMAND_INPSL;
	sendCommand(3);
}

/*
 * Sends the Jump in for DEP command
 * Equivalent to used ATR & PSL
 */
void commandInDEP() {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);

	dataBuffer[0] = PN532_COMMAND_INJUMPFORDEP;
	dataBuffer[1] = 0x00; //Passive mode
	dataBuffer[2] = 0x00; //106kbps
	dataBuffer[3] = 0x40; //NFCID3i is present

	lastCommand = PN532_COMMAND_INJUMPFORDEP;
	sendCommand(4);
}
/*
 * Send a command to retrieve the general status of the PN532
 */
void commandGeneralStatus() {
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);

	dataBuffer[0] = PN532_COMMAND_GETGENERALSTATUS;

	lastCommand = PN532_COMMAND_GETGENERALSTATUS;
	sendCommand(1);

}



////////////EXISTING CODE////////////////////////////////////////

void diagnose(alt_u8 testID, alt_u8* inParams, alt_u8 inParamsLen)
{
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);
	dataBuffer[0] = PN532_COMMAND_DIAGNOSE;
	dataBuffer[1] = testID;
	if (inParams != NULL)
		memcpy(&dataBuffer[2], inParams, inParamsLen);
	lastCommand = PN532_COMMAND_DIAGNOSE;
	sendCommand(2 + inParamsLen);
}

bool parseCommunicatonTest(alt_u8* outParams, alt_u8 inParamsLen)
{
	if (!readResponse())
		return false;

	memcpy(outParams, &dataBuffer[3], inParamsLen);
	return true;
}

void getFirmwareVersion()
{
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);
	dataBuffer[0] = PN532_COMMAND_GETFIRMWAREVERSION;
	lastCommand = PN532_COMMAND_GETFIRMWAREVERSION;
	sendCommand(1);
}

bool parseFirmwareVersion(FirmwareVersion* firmware)
{
	if (!readResponse())
		return false;

	firmware->ic = dataBuffer[2];
	firmware->ver = dataBuffer[3];
	firmware->rev = dataBuffer[4];
	firmware->support = dataBuffer[5];

	return true;
}

void samConfiguration(alt_u8 mode, alt_u8 timeout, alt_u8 irqMode)
{
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);
	dataBuffer[0] = PN532_COMMAND_SAMCONFIGURATION;
	dataBuffer[1] = mode; //Might need to use Virtual card mode with ISO14443-3, 4 type A
	dataBuffer[2] = timeout;
	dataBuffer[3] = irqMode;
	lastCommand = PN532_COMMAND_SAMCONFIGURATION;
	sendCommand(4);
}

void inListPassiveTarget(alt_u8 maxTargets, alt_u8 cardType)
{
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);
	dataBuffer[0] = PN532_COMMAND_INLISTPASSIVETARGET;
	dataBuffer[1] = maxTargets;
	dataBuffer[2] = cardType;
	dataBuffer[3] = 0;
//	dataBuffer[4] = 0x01;//probalistic polling
	lastCommand = PN532_COMMAND_INLISTPASSIVETARGET;
	sendCommand(4);
}

/*
 * Modified to use type B card target data responses
 */
bool parsePassiveTargetData(alt_u8* numTargets, TargetData* targetData)
{
	alt_u8 idLen;
	alt_u8 targetsFound;

	if (!readResponse())
		return false;

	alt_u8 testBuffer[PN532_MAX_FRAME_SIZE];
	memset(testBuffer, 0, PN532_MAX_FRAME_SIZE);
	memcpy(testBuffer, dataBuffer, PN532_MAX_FRAME_SIZE);

	targetsFound = dataBuffer[2];

	*numTargets = targetsFound;
	if (targetsFound == 1)
	{
		targetData[0].targetNumber = dataBuffer[3];
//		idLen = dataBuffer[7];
		idLen = 4; //TODO: Fix this
		targetData[0].idLen = idLen;
		memcpy(targetData[0].id, &dataBuffer[5], idLen);
	}
	else if (targetsFound == 2)
	{
		alt_u8 atsLen;
		alt_u8 target2Offset;

		targetData[0].targetNumber = dataBuffer[3];
		idLen = dataBuffer[7];
		targetData[0].idLen = idLen;
		memcpy(targetData[0].id, &dataBuffer[8], idLen);

		atsLen = dataBuffer[8 + idLen];
		target2Offset = dataBuffer[8 + idLen + atsLen];

		targetData[1].targetNumber = dataBuffer[target2Offset];
		idLen = dataBuffer[target2Offset + 4];
		targetData[1].idLen = idLen;
		memcpy(targetData[1].id, &dataBuffer[target2Offset + 5], idLen);
	}
	else
	{
		printf("Invalid number of targets\n");
		return false;
	}

	return true;
}

bool checkAck()
{
	memset(frameBuffer, 0, PN532_MAX_FRAME_SIZE);

	if (!readStatusReady())
		return false;

	I2C_ReadMore(PN532_SCL_BASE, PN532_SDA_BASE, frameBuffer, PN532_ACK_SIZE, true);
	I2C_Stop(PN532_SCL_BASE, PN532_SDA_BASE);

	alt_u8 testBuffer[PN532_MAX_FRAME_SIZE];
	memset(testBuffer, 0, PN532_MAX_FRAME_SIZE);
	memcpy(testBuffer, frameBuffer, PN532_MAX_FRAME_SIZE);

	if (memcmp(frameBuffer, ackFrame, PN532_ACK_SIZE) == 0)
		return true;
	else
		return false;
}

bool checkGeneralSuccess()
{
	return readResponse();
}

// INTERNAL FUNCTIONS BELOW //

bool readStatusReady()
{
	alt_u8 statusByte;

	I2C_Start(PN532_SCL_BASE, PN532_SDA_BASE);
//	usleep(COMM_DELAY);
	I2C_ReadFromDevice(PN532_SCL_BASE, PN532_SDA_BASE, PN532_I2C_ADDR, &statusByte, 1, false);

	//Might need to comment out this error checking
	if (statusByte == PN532_STATUS_READY)
		return true;
	else
	{
//		I2C_Stop(PN532_SCL_BASE, PN532_SDA_BASE);
//		return false;
		printf("Error: Device is not ready. Ignore and continue reading\n");
		return true;
	}
}

bool readResponse()
{
	alt_u8 dataLen;

	memset(frameBuffer, 0, PN532_MAX_FRAME_SIZE);
	memset(dataBuffer, 0, PN532_MAX_DATA_SIZE);

	if (!readStatusReady())
		return false;

	I2C_ReadMore(PN532_SCL_BASE, PN532_SDA_BASE, frameBuffer, PN532_FRAME_HEADER_SIZE, false);

	alt_u8 testBuffer[PN532_MAX_FRAME_SIZE];
	memset(testBuffer, 0, PN532_MAX_FRAME_SIZE);
	memcpy(testBuffer, frameBuffer, PN532_MAX_FRAME_SIZE);

	if (frameBuffer[0] != PN532_PREAMBLE || frameBuffer[1] != PN532_STARTCODE1 || frameBuffer[2] != PN532_STARTCODE2)
	{
		printf("Invalid preamble\n");
//		I2C_Stop(PN532_SCL_BASE, PN532_SDA_BASE);
//		return false;
	}

	dataLen = frameBuffer[3];
	if (frameBuffer[4] != (alt_u8)~dataLen + 1)
	{
		printf("Invalid length check\n");
//		I2C_Stop(PN532_SCL_BASE, PN532_SDA_BASE);
//		return false;
	}

	I2C_ReadMore(PN532_SCL_BASE, PN532_SDA_BASE, frameBuffer + PN532_FRAME_HEADER_SIZE, dataLen + PN532_FRAME_FOOTER_SIZE, true);
	I2C_Stop(PN532_SCL_BASE, PN532_SDA_BASE);

	if (frameBuffer[5] != PN532_PN532TOHOST || frameBuffer[6] != (lastCommand + 1))
	{
		printf("Invalid identifier\n");
//		return false;
	}

	if (frameBuffer[PN532_FRAME_HEADER_SIZE + dataLen + PN532_FRAME_FOOTER_SIZE - 1] != PN532_POSTAMBLE)
	{
		printf("Invalid postamble\n");
//		return false;
	}

	memcpy(dataBuffer, frameBuffer + PN532_FRAME_HEADER_SIZE, dataLen);

	memset(testBuffer, 0, PN532_MAX_FRAME_SIZE);
	memcpy(testBuffer, frameBuffer, PN532_MAX_FRAME_SIZE);

	return true;
}

void sendCommand(int cmdLen)
{
	memset(frameBuffer, 0, PN532_MAX_FRAME_SIZE);
	packFrame(cmdLen);
	sendFrame(cmdLen + PN532_FRAME_OVERHEAD);
}

void packFrame(alt_u8 cmdLen)
{
	int i;
	alt_u8 checkSum;

	frameBuffer[0] = PN532_PREAMBLE;
	frameBuffer[1] = PN532_STARTCODE1;
	frameBuffer[2] = PN532_STARTCODE2;
	frameBuffer[3] = cmdLen + 1;
	frameBuffer[4] = (alt_u8)~(cmdLen + 1) + 1;
	frameBuffer[5] = PN532_HOSTTOPN532;

	checkSum = PN532_PREAMBLE;
	checkSum += PN532_STARTCODE1;
	checkSum += PN532_STARTCODE2;
	checkSum += PN532_HOSTTOPN532;

	for (i = 0; i < cmdLen; i++)
	{
		frameBuffer[PN532_DATA_OFFSET + i] = dataBuffer[i];
		checkSum += dataBuffer[i];
	}

	frameBuffer[PN532_DATA_OFFSET + cmdLen] = (alt_u8)~checkSum;
	frameBuffer[PN532_DATA_OFFSET + cmdLen + 1] = PN532_POSTAMBLE;
}

void sendFrame(alt_u8 frmLen)
{

	I2C_Start(PN532_SCL_BASE, PN532_SDA_BASE);

	OSTimeDlyHMSM(0, 0, 0, COMM_DELAY);

	if (lastCommand == PN532_COMMAND_SAMCONFIGURATION)
	{
		if (!I2C_WriteToDevice(PN532_SCL_BASE, PN532_SDA_BASE, PN532_I2C_ADDR, frameBuffer, frmLen, 2000))
			printf("PN532 samConfig command write failed\n");
	}
	else
	{
		if (!I2C_WriteToDevice(PN532_SCL_BASE, PN532_SDA_BASE, PN532_I2C_ADDR, frameBuffer, frmLen, 0))
			printf("PN532 last command write failed\n");
	}
	I2C_Stop(PN532_SCL_BASE, PN532_SDA_BASE);
}
